/* MAC 440/5759 - Sistemas de Objetos Distribudos  */
/* IME-USP - 1o.sem/2002 - Prof. Fabio Kon          */
/*                                                  */
/* Projeto Final: Siscoma                           */
/*                                                  */
/* Emerson Luiz Navarro  - emr@linux.ime.usp.br     */
/* Leo Kazuhiro Ueda     -  lku@ime.usp.br          */
/* Pratip Roy Chowdhury  - roy@linux.ime.usp.br     */
/* Ricardo Hideo Sahara  - hideo@linux.ime.usp.br   */
/*                                                  */


/* GerenciadorDeTransferencia.java */

import java.io.*;

/**
 * Realiza o <i>download</i> de um arquivo, e oferece controle
 * e estatsticas da transferncia.
 * <p>
 * Uma instncia desta classe  associada a um arquivo, e inicia
 * uma <i>thread</i> que faz ento as requisies de pedaos
 * de um certo tamanho fixo do arquivo.
 * As requisies so feitas atraves do mtodo
 * {@link Servente#pega_arquivo(String, int, int)}.
 * <p>
 * So oferecidas operaes para interromper a transferncia e 
 * mudar o servidor de origem do arquivo.
 *  possvel tambm obter uma estimativa da taxa de transferncia.
 */
public class GerenciadorDeTransferencia {
 
  private ArquivoRemoto AR;

  private int Posicao;
  private int Bloco;

  private File F;

  private Transf T;

  private double VelPorBloco;
  private double VelMedia;

  /**
   * Cria uma instncia associada a um arquivo, e j inicia
   * a transferncia.
   *
   * @param ar o arquivo a ser transferido
   * @param tamBloco nmero de bytes que cada chamada 
   *                 a {@link Servente#pega_arquivo(String, int, int)}
   *                 vai pedir   
   */
  public GerenciadorDeTransferencia(ArquivoRemoto ar, int tamBloco) {
    AR = ar;
    Bloco = tamBloco;

    F = new File(Siscoma.DIRGRAV + "/" + ar.nome());

    if (F.length() > ar.tamanho()) throw new Error ("Arquivo j existe e  maior!");
    Posicao = (int)F.length();

    T = new Transf();
    T.start();
  }

  public ArquivoRemoto arquivoRemoto() { return AR; }

  /**
   * Interrompe (pausa) a transferncia.
   *
   * @see {@link continua}
   */
  public void interrompe() {
    T.interrompe();
    synchronized (T) {
      while (!T.interrompida()) {
	try {
	  T.wait();
	} catch (InterruptedException ie) {}
      }
    }
  }
  
  
  /**
   * Continua uma transferncia interrompida.
   *
   * @see {@link interrompe}
   */
  public void continua() {
    if (!T.interrompida()) return;

    T = new Transf();
    T.start();
  }
  
  /**
   * Troca o servente que fornece o arquivo.
   * (Interrompe a transferncia e inicia uma nova com o
   * novo servente.)
   *
   * @param ar o <code>ArquivoRemoto</code> associado ao
   *           novo servente
   */
  public void trocaServidor(ArquivoRemoto ar) 
    throws ExcArquivoDiferente
  {
    // verifica se  o mesmo arquivo
    if (ar.hash() != AR.hash() || ar.tamanho() != AR.tamanho())
      throw new ExcArquivoDiferente(AR, ar);
    interrompe();
    AR = ar;
    T = new Transf();
    T.start();
  }

  /**
   * Fornece a taxa de transferncia mdia do <i>download</i>
   * em kilobytes por segundo.
   *
   * @return a taxa de transferncia em kB/s
   */
  public double velocidadeMedia() {
    return T.velocidadeMedia()/1024;
  }
  public double velocidade() {
    return T.velocidadeMedia();
  }


  /**
   * Fornece a taxa de transferncia do <i>download</i>
   * em kilobytes por segundo, medida durante a transferncia
   * de um bloco do arquivo.
   *
   * @return a taxa de transferncia (pseudo)instantnea em kB/s
   */
  public double velocidadeInstantanea() {
    return T.velocidadeInstantanea()/1024;
  }

  /**
   * Fornece o progressso do <i>download</i>.
   * Mais precisamente, devolve o nmero de bytes recebidos dividido
   * pelo tamanho total do arquivo.
   *
   * @return a proporo do nmero de bytes recebidos em relao ao 
   *         tamanho do arquivo
   */
  public double progresso() {
    return (double)Posicao/AR.tamanho();
  }

  /**
   * Fornece o nmero de bytes recebidos.
   *
   * @return o nmero de bytes recebidos
   */
  public int bytesTransferidos() {
    return Posicao;
  }

  /**
   * Verifica se a transferncia terminou.
   *
   * @return <code>true</code> sse a transferncia terminou
   */
  public boolean terminou() { return (Posicao >= AR.tamanho()); }

  
/*  public String toString() {
    return new String(AR.nome() + "\t" + AR.nomeDoServidor() + "\t" +
		      Posicao + "/" + AR.tamanho() + "\t" +
		      T.velocidadeMedia() + "B/s");
  }
*/

  /****************************************************************************/

  class Transf extends Thread {

    boolean Interrompe;
    boolean Interrompida;
    FileOutputStream FOut;

    int posInicial = 0xDEADBEEF;
    long vmTInicial = 0xDEADBEEFL;

    //int Posicao;
    //int 
    public Transf() {
      Interrompe = false;
      Interrompida = false;
      try {
	FOut = new FileOutputStream(Siscoma.DIRGRAV + "/" + AR.nome(), true);
      }
      catch (IOException ioe) { System.err.println("IOE"); }
    }

    public synchronized void interrompe() { Interrompe = true; }
    public synchronized boolean interrompida() { 
      if (Interrompida) return true;
      if (Interrompe) {
	while (!Interrompida) {
	  try {
	    T.wait();
	  } catch (InterruptedException ie) {}
	}
	return true;
      }
      return false;
    }

    public void run() {
      System.out.println("Download: " + AR.nome() + " " + AR.tamanho());

      posInicial = Posicao;
      vmTInicial = System.currentTimeMillis();

      if (terminou()) {
	synchronized (this) {
	  Interrompida = true;
	  this.notifyAll();
	}
      }
      else {
	try {
	  while (!Interrompe && Posicao < AR.tamanho()) {
	    long viTInicial = System.currentTimeMillis();
	    int [] buf = AR.servidor().pega_arquivo(AR.nome(), Posicao, Bloco);
	    Posicao += buf.length;
	    for (int i = 0 ; i < buf.length; i++)
	      FOut.write((byte)buf[i]);
	    long viTempo = System.currentTimeMillis() - viTInicial;
	    VelPorBloco = (double)Bloco/viTempo;
	  }
	  
	  FOut.close();
	  
	}
	catch (IOException ioe) { System.err.println("IOE"); }
	System.out.println("Pos " + Posicao);
	synchronized (this) {
	  long vmTempo = System.currentTimeMillis() - vmTInicial;
	  VelMedia = (double)(Posicao-posInicial) /vmTempo;
	  Interrompida = true;
	  this.notifyAll();
	}
      }

    }

    public double velocidadeMedia() {
      if (terminou()) 
	return 1000*VelMedia;
      long vmTempo = System.currentTimeMillis() - vmTInicial;
      VelMedia = ((double)(Posicao-posInicial))/vmTempo;
      return 1000*VelMedia;
    }
    public double velocidadeInstantanea() {
      return 1000*VelPorBloco;
    }


  }
}
